"""
Bareos specific Fuse node.
"""

from   bareos.fuse.node.bvfsdir import BvfsDir
from   bareos.fuse.node.directory import Directory
from   bareos.fuse.node.file import File
from   bareos.fuse.node.joblog import JobLog
import errno
import logging
from   pprint import pformat
import stat

class Job(Directory):
    def __init__(self, root, job):
        self.job = job
        super(Job, self).__init__(root, self.get_name(job))
        try:
            if 'client' not in job:
                job['client'] = job['clientname']
        except KeyError:
            pass
        try:
            self.stat.st_ctime = self._convert_date_bareos_unix(self.job['starttime'])
        except KeyError:
            pass

        try:
            endtime = self.job['realendtime']
        except KeyError:
            pass
        if not endtime:
            try:
                endtime = self.job['endtime']
            except KeyError:
                pass
        self.stat.st_mtime = self._convert_date_bareos_unix(endtime)

        if job['jobstatus'] == 'T' or job['jobstatus'] == 'E' or job['jobstatus'] == 'W' or job['jobstatus'] == 'f':           
            self.set_static()

    @classmethod
    def get_id(cls, job):
        return job['jobid']

    def do_get_name(self, job):
        name = "jobid={jobid}_name={name}".format(**job)
        if "client" in job:
            name += "_client={client}".format(**job)
        elif "clientname" in job:
            name += "_client={clientname}".format(**job)
        if "level" in job and job['level'].strip():
            name += "_level={level}".format(**job)
        if "jobstatus" in job:
            name += "_status={jobstatus}".format(**job)
        return name

    def do_update_stat(self):
        self.subnode_count = 3

    def do_update(self):
        self.add_subnode(File, name="info.txt", content = pformat(self.job) + "\n")
        self.add_subnode(JobLog, name="job.log", job=self.job)
        if self.job['jobstatus'] != 'f':
            self.add_subnode(BvfsDir, "data", self, None)



"""
Bareos specific Fuse node.
"""

from   bareos.fuse.node.base import Base
import logging
import os
import stat

class BvfsCommon(Base):
    XattrKeyRestoreTrigger = "user.bareos.do"
    XattrKeyRestoreJobId = "user.bareos.restore_job_id"
    
    def init(self, root, jobid, path, filename, stat):
        self.jobid = jobid
        self.static = True
        self.restorepath = os.path.normpath("/%s/jobid=%s/" % (self.root.restorepath, self.jobid))
        if filename:
            self.restorepathfull = os.path.normpath("%s%s%s" % (self.restorepath, path, filename))
        else:
            self.restorepathfull = os.path.normpath("%s%s" % (self.restorepath, path))
        self.xattr = {
                'user.bareos.restorepath': self.restorepathfull,
                'user.bareos.restored': 'no',
            }
        if self.root.restoreclient:
            # restore is only possible, if a restoreclient is given
            self.xattr['user.bareos.do_options'] = 'restore'
            self.xattr['user.bareos.do'] = ''

        if stat:
            self.set_stat(stat)            

    def listxattr(self, path):
        # update xattrs
        if self.xattr['user.bareos.restored'] != 'yes':
            if os.access(self.restorepathfull, os.F_OK):
                self.xattr['user.bareos.restored'] = 'yes'
        return super(BvfsCommon, self).listxattr(path)

    # Helpers
    # =======

    def restore(self, path, pathIds, fileIds):
        self.logger.debug( "start" )
        bvfs_restore_id = "b20042"
        dirId=''
        if pathIds:
            dirId='dirid=' + ",".join(map(str, pathIds))
        fileId=''
        if fileIds:
            fileId='fileid=' + ",".join(map(str, fileIds))
        select = self.bsock.call(
            '.bvfs_restore jobid={jobid} {dirid} {fileid} path={bvfs_restore_id}'.format(
                jobid = self.jobid,
                dirid = dirId,
                fileid = fileId,
                bvfs_restore_id = bvfs_restore_id))
        restorejob=''
        if self.root.restorejob:
            restorejob="restorejob={restorejob} ".format(restorejob=self.root.restorejob)
        restore = self.bsock.call(
            'restore file=?{bvfs_restore_id} client={client} restoreclient={restoreclient} {restorejob} where="{where}" yes'.format(
                client = self.job.job['client'],
                restoreclient = self.root.restoreclient,
                restorejob = restorejob,
                where = self.restorepath,
                bvfs_restore_id = bvfs_restore_id))
        try:
            restorejobid=restore['run']['jobid']
            self.setxattr(path, self.XattrKeyRestoreJobId, str(restorejobid), 0)
        except KeyError:
            self.logger.debug("failed to get resulting jobid of run command (maybe old version of Bareos Director)")
        cleanup = self.bsock.call('.bvfs_cleanup path={bvfs_restore_id}'.format(bvfs_restore_id = bvfs_restore_id))
        self.logger.debug( "end" )



"""
Bareos-Fuse specific exceptions
"""

class ParameterMissing(Exception):
    """parameter missing"""
    pass

class SocketConnectionRefused(Exception):
    """network socket connection refused"""
    pass

class RestoreClientUnknown(Exception):
    """given restore client is not known"""
    pass

class RestoreJobUnknown(Exception):
    """given restore job is not known"""
    pass

class RestorePathInvalid(Exception):
    """restore path is invalid"""
    pass



"""
Bareos specific Fuse node.
"""

from   bareos.fuse.node.bvfscommon import BvfsCommon
from   bareos.fuse.node.file import File
import errno
import os

class BvfsFile(File, BvfsCommon):
    def __init__(self, root, file, job, bvfspath):
        super(BvfsFile, self).__init__(root, file['name'], content = None)
        self.file = file
        self.job = job
        self.bvfspath = bvfspath
        self.id = self.get_id(file, job, bvfspath)
        BvfsCommon.init(self, self.root, self.file['jobid'], self.bvfspath,  self.file['name'], self.file['stat'])

    @classmethod
    def get_id(cls, file, job, bvfspath):
        return "jobid=%s_fileid=%s" % (str(job.job['jobid']),str(file['fileid']))


    # Filesystem methods
    # ==================

    def readlink(self, path):
        self.logger.debug( "readlink %s" % (path) )
        try:
            pathname = os.readlink(self.restorepathfull)
            if pathname.startswith("/"):
                # Path name is absolute, sanitize it.
                return str(os.path.relpath(pathname, self.restorepath))
            else:
                return str(pathname)
        except OSError:
            return -errno.ENOENT

    def setxattr(self, path, key, value, flags):
        if path.len() == 0:
            if key == self.XattrKeyRestoreTrigger and value == "restore":
                if not self.root.restoreclient:
                    self.logger.error("no restoreclient given, files can not be restored")
                    return -errno.EPERM
                if not self.root.restorepath:
                    self.logger.error("no restorepath given, files can not be restored")
                    return -errno.EPERM
                else:
                    self.restore(path, None, [ self.file['fileid'] ])
        return super(BvfsFile, self).setxattr(path, key, value, flags)


    # File methods
    # ============

    def open(self, path, flags):
        return os.open(self.restorepathfull, flags)

    #def create(self, path, mode, fi=None):
        #full_path = self._full_path(path)
        #return os.open(full_path, os.O_WRONLY | os.O_CREAT, mode)

    def read(self, path, length, offset):
        try:
            fh=self.open(path, os.O_RDONLY)
            os.lseek(fh, offset, os.SEEK_SET)
            return os.read(fh, length)
        except OSError:
            return -errno.ENOENT

    #def write(self, path, buf, offset, fh):
        #os.lseek(fh, offset, os.SEEK_SET)
        #return os.write(fh, buf)

    #def truncate(self, path, length, fh=None):
        #full_path = self._full_path(path)
        #with open(full_path, 'r+') as f:
            #f.truncate(length)

    #def flush(self, path, fh):
        #return os.fsync(fh)

    #def release(self, path, fh):
        #return os.close(fh)

    #def fsync(self, path, fdatasync, fh):
        #return self.flush(path, fh)



"""
Bareos specific Fuse node.
"""

from   bareos.fuse.node.backups import Backups
from   bareos.fuse.node.directory import Directory
from   bareos.fuse.node.jobslist import JobsList

class Client(Directory):
    def __init__(self, root, name):
        super(Client, self).__init__(root, name)

    @classmethod
    def get_id(cls, name):
        return name

    def do_update(self):
        self.add_subnode(Backups, "backups", client=self.name)
        self.add_subnode(JobsList, "jobs", "client=%s" % (self.name))



"""
Bareos specific Fuse node.
"""

from   bareos.fuse.node.file import File
import stat

class VolumeStatus(File):

    __volstatus2filemode = {
        "Archive": 0o440,
        # rw
        "Append": 0o660,
        # ro
        "Full": 0o440,
        "Used": 0o440,
        # to be used
        "Recycle": 0o100,
        "Purged": 0o100,
        # not to use
        "Cleaning": 0000,
        "Error": 0000,
    }

    def __init__(self, root, basename, volume):
        super(VolumeStatus, self).__init__(root, None, None)
        self.basename = basename
        self.volume = volume
        self.do_update_stat()
        self.set_name( "%s%s" % (self.basename, self.volume['volstatus']) )

    @classmethod
    def get_id(cls, basename, volume):
        return "%s%s" % (basename, volume['mediaid'])

    def do_get_name(self, basename, volume):
        return "%s%s" % (basename, volume['volstatus'])

    def do_update(self):
        volumename = self.volume['volumename']
        data = self.bsock.call( "llist volume=%s" % (volumename) )
        self.volume = data['volume']
        self.do_update_stat()

    def do_update_stat(self):
        try:
            self.set_name( "status=%s" % (self.volume['volstatus']) )
            self.stat.st_size = int(self.volume['volbytes'])            
            self.stat.st_atime = self._convert_date_bareos_unix(self.volume['labeldate'])
            #self.stat.st_ctime = stat['ctime']
            self.stat.st_mtime = self._convert_date_bareos_unix(self.volume['lastwritten'])
        except KeyError as e:
            self.logger.warning(str(e))
            pass

        # set mode dependend on if volume is in apppend mode or full
        volstatus = self.volume['volstatus']
        if volstatus in self.__volstatus2filemode:
            self.stat.st_mode = stat.S_IFREG | self.__volstatus2filemode[volstatus]
        else:
            self.logger.warning( "volume status %s unknown" % (self.volume['volstatus']) )
            self.stat.st_mode = stat.S_IFREG | 0000



"""
Bareos specific Fuse node.
"""

from   bareos.exceptions import *
from   bareos.fuse.node.bvfscommon import BvfsCommon
from   bareos.fuse.node.bvfsfile import BvfsFile
from   bareos.fuse.node.directory import Directory
from   dateutil import parser as DateParser
import errno
import logging

class BvfsDir(Directory, BvfsCommon):
    def __init__(self, root, name, job, pathid, directory = None):
        super(BvfsDir, self).__init__(root, name)
        self.job = job
        self.jobid = job.job['jobid']
        self.pathid = pathid
        self.static = True
        self.directory = directory
        if directory:
            self.set_stat(directory['stat'])
            BvfsCommon.init(self, self.root, self.jobid, self.directory['fullpath'],  None, self.directory['stat'])

    @classmethod
    def get_id(cls, name, job, pathid, directory = None):
        id = None
        if pathid == None:
            id = "jobid=%s" % (str(job.job['jobid']))
        else:
            id = "jobid=%s_pathid=%s" % (str(job.job['jobid']),str(pathid))
        return id

    def setxattr(self, path, key, value, flags):
        if path.len() == 0:
            self.logger.debug("value: " + value + " ,type: " + str(type(value)))
            if key == self.XattrKeyRestoreTrigger and value == b"restore":
                if not self.root.restoreclient:
                    self.logger.warning("no restoreclient given, files can not be restored")
                    return -errno.ENOENT
                if not self.root.restorepath:
                    self.logger.warning("no restorepath given, files can not be restored")
                    return -errno.ENOENT
                else:
                    self.restore(path, [ self.pathid ], None)
        return super(BvfsDir, self).setxattr(path, key, value, flags)

    def do_update(self):
        directories = self.get_directories(self.pathid)
        files = self.get_files(self.pathid)
        for i in directories:
            if i['name'] != "." and i['name'] != "..":
                name = i['name'].rstrip('/')
                pathid = i['pathid']
                self.add_subnode(BvfsDir, name, self.job, pathid, i)
        self.subnode_count = len(self.subnodes)
        if self.directory:
            for i in files:
                self.add_subnode(BvfsFile, i, self.job, self.directory['fullpath'])

    def get_directories(self, pathid):
        if pathid == None:
            self.bsock.call(".bvfs_update jobid=%s" % (self.jobid))
            path = 'path=""'
        else:
            path = 'pathid=%s' % pathid
        data = self.bsock.call('.bvfs_lsdirs  jobid=%s %s' % (self.jobid, path))
        directories = data['directories']
        if pathid == None:        
            found = False
            for i in directories:
                if i['name'] != "." and i['name'] != "..":
                    if i['name'] == "/":
                        self.directory = i
                        self.pathid = i['pathid']
                        BvfsCommon.init(self, self.root, self.jobid, i['fullpath'],  None, i['stat'])
                    else:
                        found = True
            if not found:
                # skip empty root directory
                if not self.pathid:
                    # no pathid and no directoies found. This should not happen
                    return []
                return self.get_directories(self.pathid)
        return directories

    def get_files(self, pathid):
        if pathid:
            data = self.bsock.call(".bvfs_lsfiles jobid=%s pathid=%s" % (self.jobid, pathid))
            return data['files']
        else:
            return []



"""
Bareos specific Fuse node.
"""

from   bareos.fuse.node.directory import Directory
from   bareos.fuse.node.pool import Pool

class Pools(Directory):
    def __init__(self, root, name):
        super(Pools, self).__init__(root, name)
        self.static = True

    @classmethod
    def get_id(cls, name):
        return "unique"

    def do_update(self):
        data = self.bsock.call("llist pools")
        pools = data['pools']
        for i in pools:
            self.add_subnode(Pool, i['name'], i)



"""
FUSE filesystem on bareos data.
"""

from   bareos.bsock import DirectorConsoleJson
import bareos.bsock
from   bareos.util  import Path
from   bareos.fuse.root  import Root
import errno
import fuse
import logging
import socket
import stat
import os.path
from   pprint import pformat

fuse.fuse_python_api = (0, 2)


class BareosFuse(fuse.Fuse):

    def __init__(self, *args, **kw):
        self.bsock = None
        self.bareos = None
        self.restoreclient = None
        self.restorejob = None
        self.restorepath = '/var/cache/bareosfs/'
        super(BareosFuse, self).__init__(*args, **kw)
        self.multithreaded = False

    def initLogging(self):
        self.logger = logging.getLogger()
        self.logger.setLevel(logging.DEBUG)
        if hasattr(self, "logfile"):
            hdlr = logging.FileHandler(self.logfile)
            # limit message size
            formatter = logging.Formatter('%(asctime)s  %(levelname)-7s %(module)s %(funcName)s( %(message).800s )')
            hdlr.setFormatter(formatter)
            self.logger.addHandler(hdlr)

    def parse(self, *args, **kw):
        super(BareosFuse, self).parse(*args, **kw)
        if self.fuse_args.mount_expected():
            options = [ 'address', 'port', 'dirname', 'name', 'password' ]
            self.bsockParameter = {}
            for i in options:
                if hasattr(self, i):
                    self.bsockParameter[i] = getattr(self,i)
                else:
                    #self.logger.debug( "%s: missing, default: %s" %(i, str(getattr(self,i,None))))
                    pass
            if not hasattr(self, 'password'):
                raise bareos.fuse.ParameterMissing("missing parameter password")
            else:
                password = bareos.bsock.Password(self.password)
                self.bsockParameter['password']=password
            self.restorepath = os.path.normpath(self.restorepath)
            if not os.path.isabs(self.restorepath):
                raise bareos.fuse.RestorePathInvalid("restorepath must be an absolute path")


    def main(self, *args, **kw):
        # use main() instead of fsinit,
        # as this prevents FUSE from being started in case of errors.

        self.initLogging()
        self.logger.debug('start')
        if self.fuse_args.mount_expected():
            try:
                self.bsock = bareos.bsock.BSockJson(**self.bsockParameter)
            except socket.error as e:
                self.logger.exception(e)
                raise bareos.fuse.SocketConnectionRefused(e)
            self.bareos = Root(self.bsock, self.restoreclient, self.restorejob, self.restorepath)
        super(BareosFuse, self).main(*args, **kw)
        self.logger.debug('done')

    def getattr(self, path):
        if self.bareos:
            stat = self.bareos.getattr(Path(path))
            if isinstance(stat, fuse.Stat):
                self.logger.debug("{path}: dev={stat.st_dev}, ino={stat.st_ino}, mode={stat.st_mode}, nlink={stat.st_nlink}, uid={stat.st_uid}, gid={stat.st_gid}, size={stat.st_size}, atime={stat.st_atime}, ctime={stat.st_ctime}, mtime={stat.st_mtime}".format(path=path, stat=stat))
            else:
                self.logger.debug("%s: (int) %i" % (path, stat))
            return stat

    def read(self, path, size, offset):
        data = self.bareos.read(Path(path), size, offset)
        self.logger.debug("({}, {}, {}): {}".format(path, size, offset, data))
        # This should not happen.
        if len(data) > size:
            data = data[0:size-1]
        return data

    def readdir(self, path, offset):
        entries = self.bareos.readdir(Path(path), offset)
        self.logger.debug("%s: %s" % (path, entries))
        if not entries:
            return -errno.ENOENT
        else:
            return [fuse.Direntry(f) for f in entries]

    def readlink(self, path):
        result = self.bareos.readlink(Path(path))
        self.logger.debug("%s: %s" % (path, result))
        return result

    def listxattr(self, path, size):
        '''
        list extended attributes
        '''
        keys = self.bareos.listxattr(Path(path))
        if size == 0:
            # We are asked for size of the attr list, ie. joint size of attrs
            # plus null separators.
            result = len("".join(keys)) + len(keys)
            self.logger.debug("%s: len=%s" % (path, result))
        else:
            result = keys
            self.logger.debug("%s: keys=%s" % (path, ", ".join(keys)))
        return result

    def getxattr(self, path, key, size):
        '''
        get value of extended attribute
        burpfs exposes some filesystem attributes for the root directory
        (e.g. backup number, cache prefix - see FileSystem.xattr_fields_root)
        and may also expose several other attributes for each file/directory
        in the future (see FileSystem.xattr_fields)
        '''
        value = self.bareos.getxattr(Path(path), key)
        self.logger.debug("%s: %s=%s" % (path, key, str(value)))
        if value == None:
            result = -errno.ENODATA
        elif size == 0:
            # we are asked for size of the value
            result = len(value)
            self.logger.debug("%s: len=%s" % (path, result))
        else:
            result = value
            self.logger.debug("%s: %s=%s" % (path, key, result))
        return result

    def setxattr(self, path, key, value, flags):
        '''
        set value of extended attribute
        '''
        result = self.bareos.setxattr(Path(path), key, value, flags)
        self.logger.debug("%s: %s=%s: %s" % (path, key, value, str(result)))
        return result

    def rename(self, oldpath, newpath):
        result = self.bareos.rename(Path(oldpath), Path(newpath))
        self.logger.debug("%s => %s: %s" % (oldpath, newpath, result))
        return result



# ls *.py | sed -r "s/(.*)\.py/'\1', /" | tr '\n' ' '
files = ['backups',  'base',  'bvfsdir',  'bvfsfile',  'client',  'clients',  'directory',  'file',  'job',  'joblog',  'jobs',  'jobslist',  'jobsname',  'pool', 'pools', 'status', 'volume',  'volumelist', 'volumestatus']

for i in files:
    module = __import__( "bareos.fuse.node." + i, globals(), locals(), ['*'])
    for k in dir(module):
        locals()[k] = getattr(module, k)
    del k
del i
del files



"""
bareosfs root node (top level directory)
"""

import bareos.fuse.exceptions
from   bareos.fuse.nodefactory import NodeFactory
from   bareos.fuse.node.directory import Directory
from   bareos.fuse.node import *

class Root(Directory):
    """
    Define filesystem structure of root (/) directory.
    """
    def __init__(self, bsock, restoreclient, restorejob, restorepath):
        self.bsock = bsock
        self.restoreclient = restoreclient
        if restoreclient:
            data = self.bsock.call(".clients")
            if not restoreclient in [item['name'] for item in data['clients']]:
                raise bareos.fuse.RestoreClientUnknown(restoreclient)
        self.restorejob = restorejob
        if restorejob:
            data = self.bsock.call(".jobs")
            if not restorejob in [item['name'] for item in data['jobs']]:
                raise bareos.fuse.RestoreJobUnknown(restorejob)
        self.restorepath = restorepath
        super(Root, self).__init__(self, None)
        self.factory = NodeFactory(self)
        self.add_subnode(Jobs, "jobs")
        self.add_subnode(VolumeList, "volumes")
        self.add_subnode(Pools, "pools")
        self.add_subnode(Clients, "clients")
        self.add_subnode(Status, ".bareosfs-status.txt")



"""
Bareos specific Fuse node.
"""

from   bareos.fuse.node.directory import Directory
from   bareos.fuse.node.file import File
from   bareos.fuse.node.volumelist import VolumeList
from   pprint import pformat

class Pool(Directory):
    def __init__(self, root, name, pool):
        super(Pool, self).__init__(root, name)
        self.pool = pool

    @classmethod
    def get_id(cls, name, pool):
        return name

    def do_update_stat(self):
        self.subnode_count = 2

    def do_update(self):
        self.add_subnode(File, name="info.txt", content=pformat(self.pool) + "\n")
        self.add_subnode(VolumeList, name="volumes", selector="pool=%s" % (self.name))



"""
Bareos specific Fuse node.
"""

from   bareos.fuse.node.directory import Directory
from   bareos.fuse.node.job import Job

class Backups(Directory):
    def __init__(self, root, name, client):
        super(Backups, self).__init__(root, name)
        self.client = client

    @classmethod
    def get_id(cls, name, client):
        return client

    def do_update(self):
        data = self.bsock.call("llist backups client=%s" % (self.client))
        backups = data['backups']
        for i in backups:
            self.add_subnode(Job, i)



"""
NodeFactory
"""

from   bareos.fuse.node import *

class NodeFactory(object):
    """
    Creates instances of nodes, when required. directory.
    """
    def __init__(self, root):
        self.logger = logging.getLogger()
        self.root = root
        self.instances = {}

    def get_instance(self, classtype, *args, **kwargs):
        classname = classtype.__name__
        if not classname in self.instances:
            self.instances[classname] = {}
        id = classtype.get_id(*args, **kwargs)
        if not id:
            self.logger.debug(classname + ": no id defined. Creating without storing")
            instance = classtype(self.root, *args, **kwargs)
            #self.instances[classname][id] = instance
        else:
            if id in self.instances[classname]:
                self.logger.debug(classname + ": reusing " + id)
                instance = self.instances[classname][id]
            else:
                self.logger.debug(classname + ": creating " + id)
                instance = classtype(self.root, *args, **kwargs)
                self.instances[classname][id] = instance
        return instance



"""
Bareos specific Fuse node.
"""

from   bareos.fuse.node.directory import Directory
from   bareos.fuse.node.jobslist import JobsList
from   bareos.fuse.node.jobsname import JobsName

class Jobs(Directory):
    def __init__(self, root, name):
        super(Jobs, self).__init__(root, name)
        self.static = True

    @classmethod
    def get_id(cls, name):
        return "unique"

    def do_update(self):
        data = self.bsock.call(".jobs")
        jobs = data['jobs']
        for i in jobs:
            self.add_subnode(JobsName, i['name'])
        self.add_subnode(JobsList, "all")
        self.add_subnode(JobsList, "running", "jobstatus=running")
        self.add_subnode(JobsList, "each_jobname_last_run", "last")



"""
Bareos specific Fuse node.
"""

from   bareos.fuse.node.directory import Directory
from   bareos.fuse.node.volume import Volume
from   bareos.fuse.node.volumestatus import VolumeStatus

class VolumeList(Directory):
    def __init__(self, root, name, selector = 'all'):
        super(VolumeList, self).__init__(root, name)
        self.selector = selector

    @classmethod
    def get_id(cls, name, selector = 'all'):
        return selector

    def do_update(self):
        data = self.bsock.call("llist volumes %s" % (self.selector))
        volumes = data['volumes']
        for i in volumes:
            volumename = i['volumename']
            self.add_subnode(Volume, name=volumename, volume=i)
            self.add_subnode(VolumeStatus, "%s=" % (volumename), volume=i)



"""
Bareos specific Fuse node.
"""

from    bareos.fuse.node.file import File
from    datetime import datetime, timedelta


class Status(File):
    def __init__(self, root, name):
        super(Status, self).__init__(root, name)
        self.cache_timeout = timedelta(seconds=0)
        self.cache_stat_timeout = timedelta(seconds=60)
        self.objects = 0


    @classmethod
    def get_id(cls, name):
        return "unique"

    def do_update(self):
        self.objects = 0
        self.content  = b"#\n"
        self.content += b"# bareosfs cache\n"
        self.content += b"#\n"
        self.content += self.get_status(None, self.root, -1)
        self.content += b"#\n"
        self.content += b"# number objects: %i\n" % (self.objects)
        self.content += b"#\n"

    def get_status(self, name, node, indent):
        result = b""
        if name:
            result  = b" " * (indent*2)
            result += self.as_bytes(name)
            result += b"\n"
            self.objects += 1
        for i in sorted(node.subnodes.keys()):
            result += self.get_status(i, node.subnodes[i], indent+1)
        return result



#!/usr/bin/python

# -*- coding: utf-8 -*-

import bareos.fuse
import fuse
import logging
import sys

if __name__ == '__main__':
    usage = """
    Bareos Fuse filesystem: displays files from Bareos backups as a (userspace) filesystem.

    """ + fuse.Fuse.fusage

    fs = bareos.fuse.BareosFuse(
        #version="%prog: " + BAREOS_FUSE_VERSION,
        usage = usage,
        # required to let "-s" set single-threaded execution
        dash_s_do = 'setsingle',
    )

    fs.parser.add_option(
        mountopt="address",
        metavar="BAREOS_DIRECTOR",
        default='localhost',
        help="address of the Bareos Director to connect [default: \"%default\"]")
    fs.parser.add_option(
        mountopt="port",
        metavar="PORT",
        default='9101',
        help="address of the Bareos Director to connect [default: \"%default\"]")
    fs.parser.add_option(
        mountopt="dirname",
        metavar="BAREOS_DIRECTOR_NAME",
        default='',
        help="name of the Bareos Director to connect [default: \"%default\"]")
    fs.parser.add_option(
        mountopt="name",
        metavar="BAREOS_CONSOLE_NAME",
        help="name of the Bareos Named Console. If not given, it tries to use the Bareos Default Console")
    fs.parser.add_option(
        mountopt="password",
        metavar="PASSWORD",
        default='',
        help="password of the Bareos Console to authenticate at the Bareos Director")
    fs.parser.add_option(
        mountopt="restoreclient",
        metavar="BAREOS_CLIENT_NAME",
        default='',
        help="Bareos client used to restore files")
    fs.parser.add_option(
        mountopt="restorejob",
        metavar="BAREOS_JOB_NAME",
        default='',
        help="Bareos job used to restore files")
    fs.parser.add_option(
        mountopt="restorepath",
        metavar="PATH",
        default=fs.restorepath,
        help="path prefix to restore files [default: \"%default\"]")
    fs.parser.add_option(
        mountopt="logfile",
        metavar="FILENAME",
        default='',
        help="if given, log to FILENAME")

    try:
        fs.parse(values=fs, errex=1)
    except (bareos.fuse.exceptions.ParameterMissing,
            bareos.fuse.exceptions.RestorePathInvalid) as e:
        print(e.__module__ + "." + e.__class__.__name__+ ": " + str(e))
        sys.exit(1)

    try:
        fs.main()
    except (bareos.fuse.SocketConnectionRefused,
        bareos.exceptions.AuthenticationError,
        bareos.fuse.exceptions.RestoreClientUnknown,
        fuse.FuseError) as e:
            print(e.__module__ + "." + e.__class__.__name__+ ": " + str(e))
            sys.exit(1)



"""
"""

from   bareos.fuse.node.base import Base
import errno
import fuse
import logging
from   pprint import pformat
import stat

class File(Base):
    """
    File node.
    """
    def __init__(self, root, name, content = ""):
        super(File, self).__init__(root, name)
        if content is not None:
            self.content = bytes(bytearray(content, "utf-8"))
        self.stat.st_mode = stat.S_IFREG | 0o444
        self.stat.st_nlink = 1

    def get_stat(self):
        if self.content != None:
            self.stat.st_size = len(self.content)
        return super(File, self).get_stat()



"""
"""

from    datetime import datetime, timedelta
from    dateutil import parser as DateParser
import  errno
import  fuse
import  grp
import  logging
import  pwd
import  stat
import  sys

class Base(object):
    """
    Base class for filesystem nodes.
    """
    def __init__(self, root, name):
        self.logger = logging.getLogger()
        self.root = root
        self.bsock = root.bsock
        self.id = None
        self.set_name(name)
        self.stat = fuse.Stat()
        self.xattr = {}
        # content have to be an instance of "bytes" (or None).
        self.content = None
        self.subnodes = {}
        self.subnodes_old = self.subnodes.copy()
        self.subnode_count = len(self.subnodes)
        self.static = False
        self.lastupdate = None
        self.lastupdate_stat = None
        # timeout for caching
        self.cache_timeout = timedelta(seconds=60)
        self.cache_stat_timeout = timedelta(seconds=60)


    def as_str(self, data):
        if sys.version_info[0] < 3:
            # Python 2
            if isinstance(data, unicode):
                data = data.encode('utf-8')
        else:
            # Python 3
            if isinstance(data, bytes):
                data = data.decode('utf-8')
        return data


    def as_bytes(self, data):
        if data is None:
            return data
        if not isinstance(data, bytes):
            return bytes(data, "utf-8")
        return data


    @classmethod
    def get_id(cls, *args, **kwargs):
        """
        To enable reusing instances at different places in the filesystem,
        this method must be overwritten
        and return a unique id (string) for this class.
        The parameter of this must be identical to the parameter
        of the __init__ method must be identical to the parameter of get_id,
        except that the "root" parameter is excluded.
        """
        return None

    def do_get_name(self, *args, **kwargs):
        """
        Get name from init parameter.
        Normally name is statically set,
        but for some objects is it based on status data
        and dependend on where the instance is locationed 
        in the directory tree (e.g. volumestatus).
        """
        return self.name

    # Interface
    # =========

    def get_name(self, *args, **kwargs):
        result = self.do_get_name(*args, **kwargs)
        return self.as_str(result)

    def set_name(self, name):
        self.name = self.as_str(name)

    def set_static(self, value=True):
        self.static = value

    def get_stat(self):
        return self.stat

    def set_stat(self, stat):
        try:
            if stat['mode'] > 0:
                self.stat.st_mode = stat['mode']
            self.stat.st_size = stat['size']
            self.stat.st_atime = stat['atime']
            self.stat.st_ctime = stat['ctime']
            self.stat.st_mtime = stat['mtime']
        except KeyError as e:
            self.logger.warning(str(e))
            pass
        try:
            uid = pwd.getpwnam(stat['user']).pw_uid
            self.stat.st_uid = uid
        except KeyError as e:
            self.logger.info("user %s not known on this system, fall back to uid 0" % (stat['user']))
            pass
        try:
            gid = grp.getgrnam(stat['group']).gr_gid
            self.stat.st_gid = gid
        except KeyError as e:
            self.logger.info("group %s not known on this system, fall back to gid 0" % (stat['group']))
            pass
        #"stat": {
          #"atime": 1441134679,
          #"ino": 3689524,
          #"dev": 2051,
          #"mode": 33256,
          #"nlink": 1,
          #"user": "joergs",
          #"group": "joergs",
          #"ctime": 1441134679,
          #"rdev": 0,
          #"size": 1613,
          #"mtime": 1441134679
        #},

    def add_subnode(self, classtype, *args, **kwargs):
        instance = self.root.factory.get_instance(classtype, *args, **kwargs)
        name = instance.get_name(*args, **kwargs)
        if name not in self.subnodes:
            self.subnodes[name] = instance
        else:
            if name in self.subnodes_old:
                del(self.subnodes_old[name])

    def update_stat(self):
        # update status, content, ...
        now = datetime.now()
        if self.lastupdate_stat == None:
            self.logger.debug("reason: first time")
            self.do_update_stat()
            self.lastupdate_stat = now
        elif not self.static and (self.lastupdate_stat + self.cache_stat_timeout) < now:
            diff = now - self.lastupdate_stat
            self.logger.debug("reason: non-static and timeout (%d seconds)" % (diff.seconds))
            self.do_update_stat()
            self.lastupdate_stat = datetime.now()
        else:
            self.logger.debug("skipped (lastupdate: %s, static: %s)" % ( str(self.lastupdate_stat), str(self.static)))

    def do_update_stat(self):
        """
        if not overwritten, a full update is performed.
        Only updating the stat can be more efficient.
        """
        self.update()

    def update(self):
        # update status, content, ...
        now = datetime.now()
        if self.lastupdate == None:
            self.logger.debug("reason: first time")
            self.do_update()
            self.lastupdate = now
        elif not self.static and (self.lastupdate + self.cache_timeout) < now:
            diff = now - self.lastupdate
            self.logger.debug("reason: non-static and timeout (%d seconds)" % (diff.seconds))
            # store current subnodes
            #  and delete all not updated subnodes after do_update()
            self.subnodes_old = self.subnodes.copy()
            self.do_update()
            for i in list(self.subnodes_old.keys()):
                self.logger.debug("removing outdated node %s" % (i))
                del(self.subnodes[i])
            self.subnode_count = len(self.subnodes)
            self.lastupdate = datetime.now()
        else:
            self.logger.debug("skipped (lastupdate: %s, static: %s)" % ( str(self.lastupdate), str(self.static)))

    def do_update(self):
        """
        dummy, to be overwritten (not inherited!) by inherented classes
        """
        # if no node specific do_update exists,
        # remove marker for nodes to be deleted after update
        self.subnodes_old = {}


    # Filesystem methods
    # ==================

    def getattr(self, path):
        #self.logger.debug("%s(\"%s\")" % (str(self), str(path)))
        result = -errno.ENOENT
        if path.len() == 0:
            self.update_stat()
            result = self.get_stat()
        else:
            if not (path.get(0) in self.subnodes):
                self.update()
            if path.get(0) in self.subnodes:
                topdir = path.shift()
                result = self.subnodes[topdir].getattr(path)
        return result


    def read(self, path, size, offset):
        #self.logger.debug("%s(\"%s\", %d, %d)" % (str(self), str(path), size, offset))
        result = -errno.ENOENT
        if path.len() == 0:
            self.update()
            if self.content != None:
                result = self.content[offset:offset+size]
        else:
            if not (path.get(0) in self.subnodes):
                self.update()
            if path.get(0) in self.subnodes:
                topdir = path.shift()
                result = self.subnodes[topdir].read(path, size, offset)
        return result

    def readlink(self, path):
        #self.logger.debug("%s(\"%s\", %d, %d)" % (str(self), str(path), size, offset))
        result = -errno.ENOENT
        if path.len() == 0:
            pass
        else:
            if path.get(0) in self.subnodes:
                topdir = path.shift()
                result = self.subnodes[topdir].readlink(path)
        return result

    def listxattr(self, path):
        '''
        list extended attributes
        '''
        #self.logger.debug("%s(\"%s\", %s)" % (str(self.get_name()), str(path), str(size)))
        result = []
        if path.len() == 0:
            result = list(self.xattr.keys())
        else:
            if path.get(0) in self.subnodes:
                topdir = path.shift()
                result = self.subnodes[topdir].listxattr(path)
        return result

    def getxattr(self, path, key):
        '''
        get value of extended attribute
        '''
        #self.logger.debug("%s(\"%s\")" % (str(self), str(path)))
        result = None
        if path.len() == 0:
            try:
                result = self.as_str(self.xattr[key])
            except KeyError:
                pass
        else:
            if path.get(0) in self.subnodes:
                topdir = path.shift()
                result = self.subnodes[topdir].getxattr(path, key)
        return result

    def setxattr(self, path, key, value, flags):
        '''
        set value of extended attribute
        '''
        result = 0
        if path.len() == 0:
            self.xattr[key] = value
        else:
            if path.get(0) in self.subnodes:
                topdir = path.shift()
                result = self.subnodes[topdir].setxattr(path, key, value, flags)
        return result

    def rename(self, oldpath, newpath):
        self.logger.debug("%s(\"%s %s\")" % (str(self), str(oldpath), str(newpath)))
        result = -errno.ENOENT
        if path.len() == 0:
            #self.update()
            #result = self.get_stat()
            pass
        else:
            if path.get(0) in self.subnodes:
                topdir = path.shift()
                result = self.subnodes[topdir].rename(oldpath, newpath)
        return result

    # Helpers
    # =======

    def _convert_date_bareos_unix(self, bareosdate):
        unixtimestamp = 0
        try:
            unixtimestamp = int(DateParser.parse(bareosdate).strftime("%s"))
            #self.logger.debug( "unix timestamp: %d" % (unixtimestamp))
        except ValueError:
            pass
        # could happen because of timezones
        if unixtimestamp < 0:
            unixtimestamp = 0
        return unixtimestamp



"""
Bareos specific Fuse node.
"""

from   bareos.fuse.node.file import File

class JobLog(File):
    def __init__(self, root, name, job):
        super(JobLog, self).__init__(root, name)
        self.job = job
        if job['jobstatus'] == 'T' or job['jobstatus'] == 'E' or job['jobstatus'] == 'W':
            self.set_static()

    @classmethod
    def get_id(cls, name, job):
        return job['jobid']

    def do_update(self):
        jobid = self.job['jobid']
        data = self.bsock.call( "llist joblog jobid=%s" % (jobid) )
        content = ""
        for i in data['joblog']:
            content += str(i['time']) + " "
            try:
                content += str(i['logtext'])
            except KeyError:
                # some entries don't have logtext
                pass
            except UnicodeEncodeError as e:
                self.logger.error("failed to convert logtext to string: %s" % (str(e)))
                content += "BAREOFSFS SKIPPED: converting error\n"
        self.content = bytes(bytearray(content, "utf-8"))



"""
Bareos specific Fuse node.
"""

from   bareos.fuse.node.directory import Directory
from   bareos.fuse.node.job import Job

class JobsName(Directory):
    def __init__(self, root, name):
        super(JobsName, self).__init__(root, "job="+str(name))
        self.jobname=name

    @classmethod
    def get_id(cls, name):
        return name

    def do_update_stat(self):
        data = self.bsock.call("llist job=%s count" % (self.jobname))
        self.subnode_count = int(data['jobs'][0]['count'])

    def do_update(self):
        data = self.bsock.call("llist job=%s" % (self.jobname))
        jobs = data['jobs']
        for i in jobs:
            self.add_subnode(Job, i)



"""
Bareos specific Fuse node.
"""

from   bareos.fuse.node.file import File
from   bareos.fuse.node.directory import Directory
from   bareos.fuse.node.jobslist import JobsList
from   bareos.fuse.node.volumestatus import VolumeStatus
from   pprint import pformat

class Volume(Directory):
    def __init__(self, root, name, volume):
        super(Volume, self).__init__(root, name)
        self.volume = volume

    @classmethod
    def get_id(cls, name, volume):
        return volume['volumename']

    def do_update_stat(self):
        self.subnode_count = 3

    def do_update(self):
        self.add_subnode(File, name="info.txt", content=pformat(self.volume) + "\n")
        self.add_subnode(JobsList, name="jobs", selector="volume=%s" % (self.name))
        self.add_subnode(VolumeStatus, "status=", volume=self.volume)



"""
Bareos specific Fuse node.
"""

from   bareos.fuse.node.client import Client
from   bareos.fuse.node.directory import Directory

class Clients(Directory):
    def __init__(self, root, name):
        super(Clients, self).__init__(root, name)

    @classmethod
    def get_id(cls, name):
        return "unique"

    def do_update(self):
        data = self.bsock.call(".clients")
        clients = data['clients']
        for i in clients:
            name = i['name']
            self.add_subnode(Client, name)



#!/usr/bin/python

from setuptools import find_packages, setup

setup(
    name="bareos-fuse",
    #version="0.3.dev2",
    version="0.3",
    license="AGPLv3",
    author="Joerg Steffens",
    author_email="joerg.steffens@bareos.com",
    #packages=find_packages(),
    package_dir = { 
        "bareos.fuse": "bareos/fuse",
        "bareos.fuse.node": "bareos/fuse/node",
    },
    packages=["bareos.fuse", "bareos.fuse.node"],
    scripts=["bin/bareos-fuse.py"],
    url="https://github.com/bareos/bareos-fuse/",
    keywords="bareos fuse filesystem bareosfs",
    description="Virtual Filesystem showing Bareos information.",
    long_description=open("README.rst").read(),
    long_description_content_type='text/x-rst',
    install_requires=[
        "fuse-python",
        "python-bareos",
        "python-dateutil",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Topic :: System :: Archiving :: Backup",
    ],
)



"""
"""

from   bareos.fuse.node.base import Base
import errno
import logging
from   pprint import pformat
import stat

class Directory(Base):
    """
    Directory node.
    """
    def __init__(self, root, name):
        super(Directory, self).__init__(root, name)
        self.defaultdirs = [ '.', '..' ]
        self.stat.st_mode = stat.S_IFDIR | 0o755
        self.stat.st_nlink = len(self.defaultdirs)
        # arbitrary default value
        self.stat.st_size = 4096

    def readdir(self, path, offset):
        self.logger.debug("%s(\"%s\")" % (str(self.name), str(path)))
        # copy default dirs
        if path.len() == 0:
            self.update()
            result = list(self.defaultdirs) + list(self.subnodes.keys())
        else:
            if not (path.get(0) in self.subnodes):
                self.update()
            if path.get(0) in self.subnodes:
                topdir = path.shift()
                result = self.subnodes[topdir].readdir(path, offset)
        return result

    def get_stat(self):
        self.stat.st_nlink = len(self.defaultdirs) + self.subnode_count
        return super(Directory, self).get_stat()
